#include "airway-segmentation.hpp"
#include "rawFiles.hpp"
#include "openCLUtilities.hpp"
#include <iostream>
#include <queue>
#include <map>
#include <set>
#include <list>
#include <vector>
#include <math.h>

using namespace cl;

typedef struct point {
    float value;
    int x,y,z;
} point;

class PointComparison {
    public:
    bool operator() (const point &lhs, const point &rhs) const {
        return (lhs.value < rhs.value);
    }
};
struct classcomp {
      bool operator() (const int3& lhs, const int3& rhs) const
            {return lhs.x+lhs.y*1000+lhs.z*1000*1000<rhs.x+rhs.y*1000+rhs.z*1000*1000;}
};

bool inBounds(int3 pos, int3 size) {
    return pos.x > 0 && pos.y > 0 && pos.z > 0 && pos.x < size.x && pos.y < size.y && pos.z < size.z;
}

float distanceSquared(int3 pos1, int3 pos2) {
        int3 v;
        v.x= pos1.x-pos2.x;
        v.y= pos1.y-pos2.y;
        v.z= pos1.z-pos2.z;
        return v.x*v.x + v.y*v.y + v.z*v.z;
    }


int3 operator-(int3 a, int3 b) {
    int3 c;
    c.x = a.x-b.x;
    c.y = a.y-b.y;
    c.z = a.z-b.z;
    return c;
}

float dot(float3 a, int3 b) {
    return a.x*b.x+a.y*b.y+a.z*b.z;
}


// Should return a list of points that are with a certain radius from center
std::set<int3, classcomp> floodFill(int3 center, float radius, int3 size, float3 direction, float spacing = -1.0f) {
    std::set<int3, classcomp> result;
    std::stack<int3> processStack;
    processStack.push(center);
    
    while(!processStack.empty()) {
        int3 current = processStack.top();
        processStack.pop();

        if(distanceSquared(center, current) <= radius*radius && result.count(current) == 0) {
            // Additional check (cylinder)
            if(spacing > 0.0f) {
                float l = dot(direction, (current-center)) ;
                if(l > spacing)
                    continue;
            }

            result.insert(current);

            // Add all six neighbors to stack to be processed
            for(int a = -1; a < 2; a++) {
                for(int b = -1; b < 2; b++) {
                    for(int c = -1; c < 2; c++) {
                        if(a+b+c == 0)
                            continue;
                        int3 n = {current.x+a,current.y+b,current.z+c};
                        if(result.count(n) == 0 && inBounds(n, size))
                            processStack.push(n);
                    }
                }
            }
        }
    }

    return result;

};


#define LPOS(x,y,z) x+y*SIZE_X+z*SIZE_X*SIZE_Y
#define M(x,y,z) 1-sqrt(pow(T.Fx[x+y*SIZE_X+z*SIZE_X*SIZE_Y],2.0f) + pow(T.Fy[x+y*SIZE_X+z*SIZE_X*SIZE_Y],2.0f) + pow(T.Fz[x+y*SIZE_X+z*SIZE_X*SIZE_Y],2.0f))
#define SQR_MAG(pos) sqrt(pow(T.Fx[pos.x+pos.y*SIZE_X+pos.z*SIZE_X*SIZE_Y],2.0f) + pow(T.Fy[pos.x+pos.y*SIZE_X+pos.z*SIZE_X*SIZE_Y],2.0f) + pow(T.Fz[pos.x+pos.y*SIZE_X+pos.z*SIZE_X*SIZE_Y],2.0f))

#define MAX(a, b) ((a)>(b)?(a):(b))

#define SIZE 3

float hypot2(float x, float y) {
  return sqrt(x*x+y*y);
}

// Symmetric Householder reduction to tridiagonal form.

void tred2(float V[SIZE][SIZE], float d[SIZE], float e[SIZE]) {

//  This is derived from the Algol procedures tred2 by
//  Bowdler, Martin, Reinsch, and Wilkinson, Handbook for
//  Auto. Comp., Vol.ii-Linear Algebra, and the corresponding
//  Fortran subroutine in EISPACK.

  for (int j = 0; j < SIZE; j++) {
    d[j] = V[SIZE-1][j];
  }

  // Householder reduction to tridiagonal form.

  for (int i = SIZE-1; i > 0; i--) {

    // Scale to avoid under/overflow.

    float scale = 0.0f;
    float h = 0.0f;
    for (int k = 0; k < i; k++) {
      scale = scale + fabs(d[k]);
    }
    if (scale == 0.0f) {
      e[i] = d[i-1];
      for (int j = 0; j < i; j++) {
        d[j] = V[i-1][j];
        V[i][j] = 0.0f;
        V[j][i] = 0.0f;
      }
    } else {

      // Generate Householder vector.

      for (int k = 0; k < i; k++) {
        d[k] /= scale;
        h += d[k] * d[k];
      }
      float f = d[i-1];
      float g = sqrt(h);
      if (f > 0) {
        g = -g;
      }
      e[i] = scale * g;
      h = h - f * g;
      d[i-1] = f - g;
      for (int j = 0; j < i; j++) {
        e[j] = 0.0f;
      }

      // Apply similarity transformation to remaining columns.

      for (int j = 0; j < i; j++) {
        f = d[j];
        V[j][i] = f;
        g = e[j] + V[j][j] * f;
        for (int k = j+1; k <= i-1; k++) {
          g += V[k][j] * d[k];
          e[k] += V[k][j] * f;
        }
        e[j] = g;
      }
      f = 0.0f;
      for (int j = 0; j < i; j++) {
        e[j] /= h;
        f += e[j] * d[j];
      }
      float hh = f / (h + h);
      for (int j = 0; j < i; j++) {
        e[j] -= hh * d[j];
      }
      for (int j = 0; j < i; j++) {
        f = d[j];
        g = e[j];
        for (int k = j; k <= i-1; k++) {
          V[k][j] -= (f * e[k] + g * d[k]);
        }
        d[j] = V[i-1][j];
        V[i][j] = 0.0f;
      }
    }
    d[i] = h;
  }

  // Accumulate transformations.

  for (int i = 0; i < SIZE-1; i++) {
    V[SIZE-1][i] = V[i][i];
    V[i][i] = 1.0f;
    float h = d[i+1];
    if (h != 0.0f) {
      for (int k = 0; k <= i; k++) {
        d[k] = V[k][i+1] / h;
      }
      for (int j = 0; j <= i; j++) {
        float g = 0.0f;
        for (int k = 0; k <= i; k++) {
          g += V[k][i+1] * V[k][j];
        }
        for (int k = 0; k <= i; k++) {
          V[k][j] -= g * d[k];
        }
      }
    }
    for (int k = 0; k <= i; k++) {
      V[k][i+1] = 0.0f;
    }
  }
  for (int j = 0; j < SIZE; j++) {
    d[j] = V[SIZE-1][j];
    V[SIZE-1][j] = 0.0f;
  }
  V[SIZE-1][SIZE-1] = 1.0f;
  e[0] = 0.0f;
} 

// Symmetric tridiagonal QL algorithm.

void tql2(float V[SIZE][SIZE], float d[SIZE], float e[SIZE]) {

//  This is derived from the Algol procedures tql2, by
//  Bowdler, Martin, Reinsch, and Wilkinson, Handbook for
//  Auto. Comp., Vol.ii-Linear Algebra, and the corresponding
//  Fortran subroutine in EISPACK.

  for (int i = 1; i < SIZE; i++) {
    e[i-1] = e[i];
  }
  e[SIZE-1] = 0.0f;

  float f = 0.0f;
  float tst1 = 0.0f;
  float eps = pow(2.0f,-52.0f);
  for (int l = 0; l < SIZE; l++) {

    // Find small subdiagonal element

    tst1 = MAX(tst1,fabs(d[l]) + fabs(e[l]));
    int m = l;
    while (m < SIZE) {
      if (fabs(e[m]) <= eps*tst1) {
        break;
      }
      m++;
    }

    // If m == l, d[l] is an eigenvalue,
    // otherwise, iterate.

    if (m > l) {
      int iter = 0;
      do {
        iter = iter + 1;  // (Could check iteration count here.)

        // Compute implicit shift

        float g = d[l];
        float p = (d[l+1] - g) / (2.0f * e[l]);
        float r = hypot2(p,1.0f);
        if (p < 0) {
          r = -r;
        }
        d[l] = e[l] / (p + r);
        d[l+1] = e[l] * (p + r);
        float dl1 = d[l+1];
        float h = g - d[l];
        for (int i = l+2; i < SIZE; i++) {
          d[i] -= h;
        }
        f = f + h;

        // Implicit QL transformation.

        p = d[m];
        float c = 1.0f;
        float c2 = c;
        float c3 = c;
        float el1 = e[l+1];
        float s = 0.0f;
        float s2 = 0.0f;
        for (int i = m-1; i >= l; i--) {
          c3 = c2;
          c2 = c;
          s2 = s;
          g = c * e[i];
          h = c * p;
          r = hypot2(p,e[i]);
          e[i+1] = s * r;
          s = e[i] / r;
          c = p / r;
          p = c * d[i] - s * g;
          d[i+1] = h + s * (c * g + s * d[i]);

          // Accumulate transformation.

          for (int k = 0; k < SIZE; k++) {
            h = V[k][i+1];
            V[k][i+1] = s * V[k][i] + c * h;
            V[k][i] = c * V[k][i] - s * h;
          }
        }
        p = -s * s2 * c3 * el1 * e[l] / dl1;
        e[l] = s * p;
        d[l] = c * p;

        // Check for convergence.

      } while (fabs(e[l]) > eps*tst1);
    }
    d[l] = d[l] + f;
    e[l] = 0.0f;
  }
  
  // Sort eigenvalues and corresponding vectors.

  for (int i = 0; i < SIZE-1; i++) {
    int k = i;
    float p = d[i];
    for (int j = i+1; j < SIZE; j++) {
      if (fabs(d[j]) < fabs(p)) {
        k = j;
        p = d[j];
      }
    }
    if (k != i) {
      d[k] = d[i];
      d[i] = p;
      for (int j = 0; j < SIZE; j++) {
        p = V[j][i];
        V[j][i] = V[j][k];
        V[j][k] = p;
      }
    }
  }
}

void eigen_decomposition(float A[SIZE][SIZE], float V[SIZE][SIZE], float d[SIZE]) {
  float e[SIZE];
  for (int i = 0; i < SIZE; i++) {
    for (int j = 0; j < SIZE; j++) {
      V[i][j] = A[i][j];
    }
  }
  tred2(V, d, e);
  tql2(V, d, e);
}

float dot(float3 a, float3 b) {
    return a.x*b.x+a.y*b.y+a.z*b.z;
}

float3 normalize(float3 a) {
    float3 b;
    float sqrMag = sqrt(a.x*a.x+a.y*a.y+a.z*a.z);
    b.x = a.x / sqrMag;
    b.y = a.y / sqrMag;
    b.z = a.z / sqrMag;
    return b;
}

#define POS(pos) pos.x+pos.y*SIZE_X+pos.z*SIZE_X*SIZE_Y
float3 gradient(Tubeness T, int3 pos, int volumeComponent, int dimensions, int SIZE_X, int SIZE_Y, int SIZE_Z) {
    float * Fx = T.Fx;
    float * Fy = T.Fy;
    float * Fz = T.Fz;
    float f100, f_100, f010, f0_10, f001, f00_1;
    int3 npos = pos;
    switch(volumeComponent) {
        case 0:

        npos.x +=1;
        f100 = Fx[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
        npos.x -=2;
        f_100 = Fx[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
        if(dimensions > 1) {
            npos = pos;
            npos.y += 1;
            f010 = Fx[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
            npos.y -= 2;
            f0_10 = Fx[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
        }
        if(dimensions > 2) {
            npos = pos;
            npos.z += 1;
            f001 = Fx[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
            npos.z -= 2;
            f00_1 =Fx[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
        }
    break;
        case 1:

        npos.x +=1;
        f100 = Fy[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
        npos.x -=2;
        f_100 = Fy[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
        if(dimensions > 1) {
            npos = pos;
            npos.y += 1;
            f010 = Fy[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
            npos.y -= 2;
            f0_10 = Fy[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
        }
        if(dimensions > 2) {
            npos = pos;
            npos.z += 1;
            f001 = Fy[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
            npos.z -= 2;
            f00_1 =Fy[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
        }
    break;
        case 2:

        npos.x +=1;
        f100 = Fz[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
        npos.x -=2;
        f_100 = Fz[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
        if(dimensions > 1) {
            npos = pos;
            npos.y += 1;
            f010 = Fz[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
            npos.y -= 2;
            f0_10 = Fz[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
        }
        if(dimensions > 2) {
            npos = pos;
            npos.z += 1;
            f001 = Fz[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
            npos.z -= 2;
            f00_1 =Fz[POS(npos)]/sqrt(Fx[POS(npos)]*Fx[POS(npos)]+Fy[POS(npos)]*Fy[POS(npos)]+Fz[POS(npos)]*Fz[POS(npos)]);
        }
    break;
    }

    float3 grad = {
        0.5f*(f100-f_100), 
        0.5f*(f010-f0_10),
        0.5f*(f001-f00_1)
    };


    return grad;
}

float sign(float a) {
    return a < 0 ? -1.0f: 1.0f;
}

float3 getTubeDirection(Tubeness T, int3 pos,int SIZE_X,int SIZE_Y,int SIZE_Z) {

    // Do gradient on Fx, Fy and Fz and normalization
    float3 Fx = gradient(T, pos,0,1,SIZE_X,SIZE_Y,SIZE_Z);
    float3 Fy = gradient(T, pos,1,2,SIZE_X,SIZE_Y,SIZE_Z);
    float3 Fz = gradient(T, pos,2,3,SIZE_X,SIZE_Y,SIZE_Z);
    
    float Hessian[3][3] = {
        {Fx.x, Fy.x, Fz.x},
        {Fy.x, Fy.y, Fz.y},
        {Fz.x, Fz.y, Fz.z}
    };
    /*
    std::cout << "Hessian:" <<std::endl;
    std::cout << Fx.x << " " << Fy.x << " " << Fz.x << std::endl;
    std::cout << Fy.x << " " << Fy.y << " " << Fz.y << std::endl;
    std::cout << Fz.x << " " << Fz.y << " " << Fz.z << std::endl;
    */
    float eigenValues[3];
    float eigenVectors[3][3];
    eigen_decomposition(Hessian, eigenVectors, eigenValues);
    float3 e1 = {eigenVectors[0][0], eigenVectors[1][0], eigenVectors[2][0]};
    return e1;
}

void doEigen(Tubeness T, int3 pos,int SIZE_X,int SIZE_Y,int SIZE_Z, float3 * lambda, float3 * e1, float3 * e2, float3 * e3) {

    // Do gradient on Fx, Fy and Fz and normalization
    float3 Fx = gradient(T, pos,0,1,SIZE_X,SIZE_Y,SIZE_Z);
    float3 Fy = gradient(T, pos,1,2,SIZE_X,SIZE_Y,SIZE_Z);
    float3 Fz = gradient(T, pos,2,3,SIZE_X,SIZE_Y,SIZE_Z);
    
    float Hessian[3][3] = {
        {Fx.x, Fy.x, Fz.x},
        {Fy.x, Fy.y, Fz.y},
        {Fz.x, Fz.y, Fz.z}
    };
    float eigenValues[3];
    float eigenVectors[3][3];
    eigen_decomposition(Hessian, eigenVectors, eigenValues);
    *e1 = {eigenVectors[0][0], eigenVectors[1][0], eigenVectors[2][0]};
    *e2 = {eigenVectors[0][1], eigenVectors[1][1], eigenVectors[2][1]};
    *e3 = {eigenVectors[0][2], eigenVectors[1][2], eigenVectors[2][2]};
    *lambda = {eigenValues[0], eigenValues[1], eigenValues[2]};
}

int * runCenterlineExtraction(Tubeness T, float Thigh, int SIZE_X, int SIZE_Y, int SIZE_Z, std::stack<CenterlinePoint> * centerlineStack) {

    int Dmin = 7;
    float Mlow = 0.2f;
    float Tlow = 0.4f;
    int maxBelowTlow = 3;
    float minMeanTube = 0.6f;
    int TreeMin = 200;

    int * centerlines = new int[SIZE_X*SIZE_Y*SIZE_Z];

    // Create queue
    std::priority_queue<point, std::vector<point>, PointComparison> queue;
    int3 size = {SIZE_X,SIZE_Y,SIZE_Z};
    
    // Collect all valid start points
    for(int z = 1; z < SIZE_Z-1; z++) {
        for(int y = 1; y < SIZE_Y-1; y++) {
            for(int x = 1; x < SIZE_X-1; x++) {
                if(T.value[LPOS(x,y,z)] < Thigh)
                    continue;

                    int3 center = {x,y,z};
                    float radius = T.radius[LPOS(x,y,z)];
                    //if(radius < 3.0f)
                        radius = 3.0f;

                    float spacing = 3.0f;

                    float3 direction = getTubeDirection(T, center, size.x, size.y, size.z); 
                    std::set<int3, classcomp> neighbors = floodFill(center, radius, size, direction, spacing);
                    std::set<int3, classcomp>::iterator it;
                    bool valid = true;
                    for(it = neighbors.begin(); it != neighbors.end(); it++) {
                        if(T.value[LPOS(it->x, it->y, it->z)] > T.value[LPOS(x,y,z)]) {
                            valid = false;
                            break;
                        }
                    }

                if(valid) {
                    point p;
                    p.value = T.value[LPOS(x,y,z)];
                    p.x = x;
                    p.y = y;
                    p.z = z;
                    queue.push(p);
                }
            }
        }
    }


    std::cout << "Processing " << queue.size() << " valid start points" << std::endl;
    int counter = 1;
    T.value[0] = 0;
    T.Fx[0] = 1;
    T.Fy[0] = 0;
    T.Fz[0] = 0;


    // Create a map of centerline distances
    std::map<int, int> centerlineDistances;

    // Create a map of centerline stacks
    std::map<int, std::stack<CenterlinePoint> > centerlineStacks;

    while(!queue.empty()) {
        // Traverse from new start point
        point p = queue.top();
        queue.pop();

        // Has it been handled before?
        if(centerlines[LPOS(p.x,p.y,p.z)] == 1)
            continue;

        char * newCenterlines = new char[SIZE_X*SIZE_Y*SIZE_Z];
        newCenterlines[LPOS(p.x,p.y,p.z)] = 1;
        int distance = 1;
        int connections = 0;
        int prevConnection = -1;
        int secondConnection = -1;
        float meanTube = T.value[LPOS(p.x,p.y,p.z)];

        // Create new stack for this centerline
        std::stack<CenterlinePoint> stack;

        // For each direction
        for(int direction = -1; direction < 3; direction += 2) {
            int belowTlow = 0;
            int3 position = {p.x,p.y,p.z};
            float3 t_i = getTubeDirection(T, position, SIZE_X,SIZE_Y,SIZE_Z);
            t_i.x *= direction;
            t_i.y *= direction;
            t_i.z *= direction;
            float3 t_i_1;
            t_i_1.x = t_i.x;
            t_i_1.y = t_i.y;
            t_i_1.z = t_i.z;


            // Traverse
            while(true) {
                int3 maxPoint = {0,0,0};

                // Check for out of bounds
                if(position.x < 2 || position.x > SIZE_X-2 || position.y < 2 || position.y > SIZE_Y-2 || position.z < 2 || position.z > SIZE_Z-2)
                    break;

                // Try to find next point from all neighbors
                for(int a = -1; a < 2; a++) {
                    for(int b = -1; b < 2; b++) {
                        for(int c = -1; c < 2; c++) {
                            int3 n = {position.x+a,position.y+b,position.z+c};
                            if((a | b | c) == 0 || T.value[POS(n)] == 0.0f)
                                continue;

                            float3 dir = {n.x-position.x,n.y-position.y,n.z-position.z};
                            dir = normalize(dir);
                            if( (dir.x*t_i.x+dir.y*t_i.y+dir.z*t_i.z) <= 0.1)
                                continue;

                            if(T.radius[POS(n)] >= 1.5f) {
                                if(M(n.x,n.y,n.z) > M(maxPoint.x,maxPoint.y,maxPoint.z)) 
                                maxPoint = n;
                            } else {
                                if(T.value[LPOS(n.x,n.y,n.z)]*M(n.x,n.y,n.z) > T.value[POS(maxPoint)]*M(maxPoint.x,maxPoint.y,maxPoint.z)) 
                                maxPoint = n;
                            }

                        }
                    }
                }

                if(maxPoint.x+maxPoint.y+maxPoint.z > 0) {
                    // New maxpoint found, check it!
                    if(centerlines[LPOS(maxPoint.x,maxPoint.y,maxPoint.z)] > 0) {
                        // Hit an existing centerline
                        if(prevConnection == -1) {
                            prevConnection = centerlines[LPOS(maxPoint.x,maxPoint.y,maxPoint.z)];
                        } else {
                            if(prevConnection ==centerlines[LPOS(maxPoint.x,maxPoint.y,maxPoint.z)]) {
                                // A loop has occured, reject this centerline
                                connections = 5;
                            } else {
                                secondConnection = centerlines[LPOS(maxPoint.x,maxPoint.y,maxPoint.z)];
                            }
                        }
                        break;
                    } else if(M(maxPoint.x,maxPoint.y,maxPoint.z) < Mlow || (belowTlow > maxBelowTlow && T.value[LPOS(maxPoint.x,maxPoint.y,maxPoint.z)] < Tlow)) {
                        // New point is below thresholds
                        break;
                    } else if(newCenterlines[LPOS(maxPoint.x,maxPoint.y,maxPoint.z)] == 1) {
                        // Loop detected!
                        break;
                    } else {
                        // Point is OK, proceed to add it and continue
                        if(T.value[LPOS(maxPoint.x,maxPoint.y,maxPoint.z)] < Tlow) {
                            belowTlow++;
                        } else {
                            belowTlow = 0;
                        }

                        // Update direction
                        //float3 e1 = getTubeDirection(T, maxPoint,SIZE_X,SIZE_Y,SIZE_Z);

                        //TODO: check if all eigenvalues are negative, if so find the egeinvector that best matches
                        float3 lambda, e1, e2, e3;
                        doEigen(T, maxPoint, SIZE_X, SIZE_Y, SIZE_Z, &lambda, &e1, &e2, &e3);
                        if((lambda.x < 0 && lambda.y < 0 && lambda.z < 0)) {
                            if(fabs(dot(t_i, e3)) > fabs(dot(t_i, e2))) {
                                if(fabs(dot(t_i, e3)) > fabs(dot(t_i, e1))) {
                                    e1 = e3;
                                }
                            } else if(fabs(dot(t_i, e2)) > fabs(dot(t_i, e1))) {
                                e1 = e2;
                            }
                        }


                        float maintain_dir = sign(dot(e1,t_i));
                        float3 vec_sum; 
                        vec_sum.x = maintain_dir*e1.x + t_i.x + t_i_1.x;
                        vec_sum.y = maintain_dir*e1.y + t_i.y + t_i_1.y;
                        vec_sum.z = maintain_dir*e1.z + t_i.z + t_i_1.z;
                        vec_sum = normalize(vec_sum);
                        t_i_1 = t_i;
                        t_i = vec_sum;

                        // update position
                        position = maxPoint;
                        distance ++;
                        newCenterlines[LPOS(maxPoint.x,maxPoint.y,maxPoint.z)] = 1;
                        meanTube += T.value[LPOS(maxPoint.x,maxPoint.y,maxPoint.z)];

                        // Create centerline point
                        CenterlinePoint p;
                        p.pos = position;
                        p.next = &(stack.top()); // add previous 

                        // Add point to stack
                        stack.push(p);
                    }
                } else {
                    // No maxpoint found, stop!
                    break;
                }

            } // End traversal
        } // End for each direction

        // Check to see if new traversal can be added
        //std::cout << "Finished. Distance " << distance << " meanTube: " << meanTube/distance << std::endl;
        if(distance > Dmin && meanTube/distance > minMeanTube && connections < 2) {
            std::cout << "Finished. Distance " << distance << " meanTube: " << meanTube/distance << std::endl;
            std::cout << "------------------- New centerlines added #" << counter << " -------------------------" << std::endl;

            // TODO add it
            if(prevConnection == -1) {
                // No connections
                for(int i = 0; i < SIZE_X*SIZE_Y*SIZE_Z;i++) {
                    if(newCenterlines[i] > 0)
                        centerlines[i] = counter;
                }
                centerlineDistances[counter] = distance;
                centerlineStacks[counter] = stack;
                counter ++;
            } else {
                // The first connection

                std::stack<CenterlinePoint> prevConnectionStack = centerlineStacks[prevConnection];
                while(!stack.empty()) {
                    prevConnectionStack.push(stack.top());
                    stack.pop();
                }

                for(int i = 0; i < SIZE_X*SIZE_Y*SIZE_Z;i++) {
                    if(newCenterlines[i] > 0)
                        centerlines[i] = prevConnection;
                }
                centerlineDistances[prevConnection] += distance;
                if(secondConnection != -1) {
                    // Two connections, move secondConnection to prevConnection
                    std::stack<CenterlinePoint> secondConnectionStack = centerlineStacks[secondConnection];
                    centerlineStacks.erase(secondConnection);
                    while(!secondConnectionStack.empty()) {
                        prevConnectionStack.push(secondConnectionStack.top());
                        secondConnectionStack.pop();
                    }

                    for(int i = 0; i < SIZE_X*SIZE_Y*SIZE_Z;i++) {
                        if(centerlines[i] == secondConnection)
                            centerlines[i] = prevConnection;
                    }
                    centerlineDistances[prevConnection] += centerlineDistances[secondConnection];
                    centerlineDistances.erase(secondConnection);
                }

                centerlineStacks[prevConnection] = prevConnectionStack;
            }
        } // end if new point can be added
    } // End while queue is not empty


    // Find largest connected tree and all trees above a certain size
    std::map<int, int>::iterator it;
    int max = centerlineDistances.begin()->first;
    std::list<int> trees;
    for(it = centerlineDistances.begin(); it != centerlineDistances.end(); it++) {
        if(it->second > centerlineDistances[max])
            max = it->first;
        if(it->second > TreeMin) 
            trees.push_back(it->first);
    }
    std::list<int>::iterator it2;
    // TODO: if use the method with TreeMin have to add them to centerlineStack also
    *centerlineStack = centerlineStacks[max];
    for(it2 = trees.begin(); it2 != trees.end(); it2++) {
        while(!centerlineStacks[*it2].empty()) {
            centerlineStack->push(centerlineStacks[*it2].top());
            centerlineStacks[*it2].pop();
        }
    }

    // Mark largest tree with 1, and rest with 0
    for(int i = 0; i < SIZE_X*SIZE_Y*SIZE_Z;i++) {
        if(centerlines[i] == max) {
        //if(centerlines[i] > 0) {
            centerlines[i] = 1;
        } else {
            bool valid = false;
            for(it2 = trees.begin(); it2 != trees.end(); it2++) {
                if(centerlines[i] == *it2) {
                    centerlines[i] = 1;
                    valid = true;
                    break;
                }
            }
            if(!valid)
                centerlines[i] = 0;
                    
        }
    }

    return centerlines;
}

char * runCircleSegmentation(Tubeness T, std::stack<CenterlinePoint> centerline, int SIZE_X, int SIZE_Y, int SIZE_Z) {
    char * segmentation = new char[SIZE_X*SIZE_Y*SIZE_Z]();
    while(!centerline.empty()) {
        CenterlinePoint p = centerline.top();
        centerline.pop();
        float3 lambda, e1, e2, e3;
        float radius = T.radius[POS(p.pos)];
        doEigen(T, p.pos, SIZE_X, SIZE_Y, SIZE_Z, &lambda, &e1, &e2, &e3);
        for(int sample = 0; sample < 100; sample++) {
            float alpha = 2.0f*M_PI * (float)sample / 100.0f;
            float3 V_alpha;
            V_alpha.x = cos(alpha)*e3.x + sin(alpha)*e2.x;
            V_alpha.y = cos(alpha)*e3.y + sin(alpha)*e2.y;
            V_alpha.z = cos(alpha)*e3.z + sin(alpha)*e2.z;
            float3 position;
            position.x = p.pos.x + radius*V_alpha.x;
            position.y = p.pos.y + radius*V_alpha.y;
            position.z = p.pos.z + radius*V_alpha.z;
            int3 intPosition;
            intPosition.x = round(position.x);
            intPosition.y = round(position.y);
            intPosition.z = round(position.z);
            segmentation[POS(intPosition)] = 1;
        }
    }

    return segmentation;
}



char * runInverseGradientSegmentation(Tubeness T, std::stack<CenterlinePoint> centerline, int SIZE_X, int SIZE_Y, int SIZE_Z) {
    char * segmentation = new char[SIZE_X*SIZE_Y*SIZE_Z]();

    // Dilate centerline
    std::stack<int3> voxelStack;
    std::cout << "Size of centerline stack is: " << centerline.size() << std::endl;
    while(!centerline.empty()) {
        CenterlinePoint p = centerline.top();
        centerline.pop();
        // Add neighbors of current point
        segmentation[POS(p.pos)] = 1;
        int size = 1; // 3x3
        /*
        if(T.radius[POS(p.pos)] > 2.0f) 
            size = 2;
        if(T.radius[POS(p.pos)] > 4.5f) 
            size = 3;
        if(T.radius[POS(p.pos)] > 6.0f)
            size = 4;
        */

        for(int a = -size; a < size+1; a++) {
        for(int b = -size; b < size+1; b++) {
        for(int c = -size; c < size+1; c++) {
            int3 n;
            //if(size == 1 && abs(a) + abs(b) + abs(c) > 1.5f)
            //    continue;
            n.x = p.pos.x + a;
            n.y = p.pos.y + b;
            n.z = p.pos.z + c;
            if(n.x > 0 && n.y > 0 && n.z > 0 && n.x < SIZE_X && n.y < SIZE_Y && n.z < SIZE_Z) {
                if(segmentation[POS(n)] == 0) {
                    if(abs(a)+abs(b)+abs(c) < 1.5f)
                        segmentation[POS(n)] = 1;
                    voxelStack.push(n);
                }
            }
        }}}
    }


    // Run inverse Gradient segmentation on the dilated centerline
    while(!voxelStack.empty()) {
        int3 X = voxelStack.top();
        voxelStack.pop();
        //if(segmentation[POS(X)] == 1)
        //    continue;

        float maxProduct = -2.0f;
        float maxProductInverse = -2.0f;
        int3 maxN;
        int3 maxIN;
        float sqrMag = SQR_MAG(X);
        float3 FN;
        FN.x = T.Fx[POS(X)] / sqrMag;
        FN.y = T.Fy[POS(X)] / sqrMag;
        FN.z = T.Fz[POS(X)] / sqrMag;

        for(int a = -1; a < 2; a++) {
        for(int b = -1; b < 2; b++) {
        for(int c = -1; c < 2; c++) {
            if(a+b+c == 0)
                continue;

            int3 Y;
            Y.x = X.x + a;
            Y.y = X.y + b;
            Y.z = X.z + c;
            float3 XY;
            XY.x = Y.x - X.x;
            XY.y = Y.y - X.y;
            XY.z = Y.z - X.z;
            float3 YX;
            YX.x = X.x - Y.x;
            YX.y = X.y - Y.y;
            YX.z = X.z - Y.z;
            XY = normalize(XY);
            YX = normalize(YX);

            // Find the neighbors the vectors are pointing to
            if(dot(FN, XY) > maxProduct) {
                maxN = Y;
                maxProduct = dot(FN, XY);
            }
            if(dot(FN, YX) > maxProductInverse) {
                maxIN = Y;
                maxProductInverse = dot(FN, YX);
            }

            

            /*
            // Check out of bounds
            if(Y.x > 0 && Y.y > 0 && Y.z > 0 && Y.x < SIZE_X && 
                    Y.y < SIZE_Y && Y.z < SIZE_Z && SQR_MAG(Y) > SQR_MAG(X)
                    && segmentation[POS(Y)] == 0) {

                int3 Z;
                float maxDotProduct = -2.0f;
                float sqrMag = SQR_MAG(Y);
                float3 FN;
                FN.x = T.Fx[POS(Y)] / sqrMag;
                FN.y = T.Fy[POS(Y)] / sqrMag;
                FN.z = T.Fz[POS(Y)] / sqrMag;
                for(int a2 = -1; a2 < 2; a2++) {
                for(int b2 = -1; b2 < 2; b2++) {
                for(int c2 = -1; c2 < 2; c2++) {
                    if(a2+b2+c2 == 0)
                        continue;
                    int3 Zc;
                    Zc.x = Y.x+a2;
                    Zc.y = Y.y+b2;
                    Zc.z = Y.z+c2;
                    float3 YZ;
                    YZ.x = Zc.x-Y.x;
                    YZ.y = Zc.y-Y.y;
                    YZ.z = Zc.z-Y.z;
                    //YZ = normalize(YZ);
                    if(dot(FN, YZ) > maxDotProduct) {
                        maxDotProduct = dot(FN, YZ);
                        Z = Zc;
                    }
                }}}

                if(Z.x == X.x && Z.y == X.y && Z.z == X.z) {
                    segmentation[POS(X)] = 1;
                    voxelStack.push(Y);
                }
            }
            */

            /*
            // Check out of bounds
            if(Y.x > 0 && Y.y > 0 && Y.z > 0 && Y.x < SIZE_X && 
                    Y.y < SIZE_Y && Y.z < SIZE_Z && SQR_MAG(Y) > SQR_MAG(X)) {
                float3 N;
                N.x = X.x-Y.x;
                N.y = X.y-Y.y;
                N.z = X.z-Y.z;
                N = normalize(N);
                float sqrMag = SQR_MAG(Y);
                float3 FN;
                FN.x = T.Fx[POS(Y)] / sqrMag;
                FN.y = T.Fy[POS(Y)] / sqrMag;
                FN.z = T.Fz[POS(Y)] / sqrMag;
                if(FN.x*N.x + FN.y*N.y + FN.z*N.z >= 0.9f) {
                    segmentation[POS(X)] = 1;
                    if(segmentation[POS(Y)] == 0) {
                        voxelStack.push(Y);
                    }
                }
            }
            */
            /*
            if(Y.x > 0 && Y.y > 0 && Y.z > 0 && Y.x < SIZE_X && 
                    Y.y < SIZE_Y && Y.z < SIZE_Z && SQR_MAG(Y) > SQR_MAG(X)) {
                float3 N;
                N.x = X.x-Y.x;
                N.y = X.y-Y.y;
                N.z = X.z-Y.z;
                N = normalize(N);
                float3 FY, FX;
                float sqrMagX = SQR_MAG(X);
                float sqrMagY = SQR_MAG(Y);
                FX.x = T.Fx[POS(X)] / sqrMagX;
                FX.y = T.Fy[POS(X)] / sqrMagX;
                FX.z = T.Fz[POS(X)] / sqrMagX;
                FY.x = T.Fx[POS(Y)] / sqrMagY;
                FY.y = T.Fy[POS(Y)] / sqrMagY;
                FY.z = T.Fz[POS(Y)] / sqrMagY;
                bool valid = false;
                if(dot(FX, FY) > 0.75) {
                    if(dot(N, FX) > 0.0f) {
                        // X is in front of Y
                        if(sqrMagY > sqrMagX)
                            valid = true;
                    }else{
                        // X is behind Y
                        if(sqrMagX > sqrMagY)
                            valid = true;
                    }
                }

                if(valid) {
                    segmentation[POS(X)] = 1;
                    if(segmentation[POS(Y)] == 0)
                        voxelStack.push(Y);
                }
            }
            */
                        

        }}}

        if(sqrMag > SQR_MAG(maxN) && segmentation[POS(maxN)] == 1 ) {
            // accept X
            segmentation[POS(X)] = 1;
        }
        // Add neighbors
        if(segmentation[POS(X)] == 1) {
            for(int a = -1; a < 2; a++) {
            for(int b = -1; b < 2; b++) {
            for(int c = -1; c < 2; c++) {
                if(a+b+c == 0)
                    continue;

                int3 Y;
                Y.x = X.x + a;
                Y.y = X.y + b;
                Y.z = X.z + c;
                if(Y.x > 0 && Y.y > 0 && Y.z > 0 && Y.x < SIZE_X && 
                    Y.y < SIZE_Y && Y.z < SIZE_Z) {
                    if(segmentation[POS(Y)]==0){
                        voxelStack.push(Y);
                    }
                }
            }}}
        }


            
    }


    return segmentation;
}


Tubeness runCircleFittingTDF(short * HUvolume, float Fmax, float sigma, float rMin, float rMax, int SIZE_X, int SIZE_Y, int SIZE_Z, int HUthreshold, int HUminimum, int ITERATIONS, float MU) {
    Tubeness T;

    try {
        Context context = createCLContext(CL_DEVICE_TYPE_CPU);

        // Get a list of devices
        vector<Device> devices = context.getInfo<CL_CONTEXT_DEVICES>();
        std::cout << "Using device: " << devices[0].getInfo<CL_DEVICE_NAME>() << std::endl;

        // Create a command queue and use the first device
        CommandQueue queue = CommandQueue(context, devices[0], CL_QUEUE_PROFILING_ENABLE);

        // Query the size of available memory
        unsigned int memorySize = devices[0].getInfo<CL_DEVICE_GLOBAL_MEM_SIZE>();
        std::cout << "Available memory on selected device " << (double)memorySize/(1024*1024) << " MB "<< std::endl;

        // Compile source code
        Program program = buildProgramFromSource(context, "kernels.cl");

        // Create kernels
        Kernel preProcessKernel(program, "preProcess");
        Kernel blurVolumeWithGaussianKernel(program, "blurVolumeWithGaussian");
        Kernel createVectorFieldKernel(program, "createVectorField");
        Kernel circleFittingTDFKernel(program, "circleFittingTDF");

        cl::size_t<3> offset;
        offset[0] = 0;
        offset[1] = 0;
        offset[2] = 0;
        cl::size_t<3> region;
        region[0] = SIZE_X;
        region[1] = SIZE_Y;
        region[2] = SIZE_Z;

        // Transfer raw data to device as an image
        Image3D imageHUvolume = Image3D(context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, ImageFormat(CL_R, CL_SIGNED_INT16), SIZE_X, SIZE_Y, SIZE_Z, 0, 0, HUvolume);
        Buffer processedVolume = Buffer(context, CL_MEM_WRITE_ONLY, sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z);
        delete[] HUvolume;
        
        std::cout << "Input data transferred to device." << std::endl;

        std::cout << "Running TDF ..." << std::endl;

        // Run preProcess kernel
        preProcessKernel.setArg(0, imageHUvolume);
        preProcessKernel.setArg(1, processedVolume);
        preProcessKernel.setArg(2, HUthreshold);
        preProcessKernel.setArg(3, HUminimum);
        queue.enqueueNDRangeKernel(
                preProcessKernel,
                NullRange,
                NDRange(SIZE_X,SIZE_Y,SIZE_Z),
                NullRange
        );

        // Put processedVolume buffer into image
        Image3D imageVolume = Image3D(context, CL_MEM_READ_ONLY, ImageFormat(CL_R, CL_FLOAT), SIZE_X, SIZE_Y, SIZE_Z);
        queue.enqueueCopyBufferToImage(processedVolume, imageVolume, 0, offset, region);
        Buffer blurredVolume = Buffer(context, CL_MEM_WRITE_ONLY, sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z);

        // Run blurVolumeWithGaussian on processedVolume
        blurVolumeWithGaussianKernel.setArg(0, imageVolume);
        blurVolumeWithGaussianKernel.setArg(1, blurredVolume);
        blurVolumeWithGaussianKernel.setArg(2, 0.5f);
        queue.enqueueNDRangeKernel(
                blurVolumeWithGaussianKernel,
                NullRange,
                NDRange(SIZE_X,SIZE_Y,SIZE_Z),
                NullRange
        );
    
        // Copy buffer to image
        queue.enqueueCopyBufferToImage(blurredVolume, imageVolume, 0, offset, region);

        Buffer vectorField = Buffer(context, CL_MEM_WRITE_ONLY, sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z*4);
        
        // Run create vector field
        createVectorFieldKernel.setArg(0, imageVolume);
        createVectorFieldKernel.setArg(1, vectorField);
        createVectorFieldKernel.setArg(2, Fmax);

        queue.enqueueNDRangeKernel(
                createVectorFieldKernel,
                NullRange,
                NDRange(SIZE_X,SIZE_Y,SIZE_Z),
                NullRange
        );
        
        // Copy buffer to image
        Image3D iVectorField = Image3D(context, CL_MEM_READ_ONLY, ImageFormat(CL_RGBA, CL_FLOAT), SIZE_X, SIZE_Y, SIZE_Z);
        queue.enqueueCopyBufferToImage(vectorField, iVectorField, 0, offset, region);

        // Run circle fitting TDF kernel
        Buffer bufferT = Buffer(context, CL_MEM_WRITE_ONLY, sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z);
        Buffer bufferRadius = Buffer(context, CL_MEM_WRITE_ONLY, sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z);
        circleFittingTDFKernel.setArg(0, iVectorField);
        circleFittingTDFKernel.setArg(1, bufferT);
        circleFittingTDFKernel.setArg(2, bufferRadius);
        circleFittingTDFKernel.setArg(3, rMin);
        circleFittingTDFKernel.setArg(4, 3.0f);
        circleFittingTDFKernel.setArg(5, 0.5f);

        queue.enqueueNDRangeKernel(
                circleFittingTDFKernel,
                NullRange,
                NDRange(SIZE_X,SIZE_Y,SIZE_Z),
                NullRange
        );
        // Transfer buffer back to host
        float * Tsmall = new float[SIZE_X*SIZE_Y*SIZE_Z];
        float * Radiussmall = new float[SIZE_X*SIZE_Y*SIZE_Z];
        queue.enqueueReadBuffer(bufferT, CL_TRUE, 0, sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z, Tsmall);
        queue.enqueueReadBuffer(bufferRadius, CL_TRUE, 0, sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z, Radiussmall);
        /* Large Airways */
        
        // Put processedVolume buffer into image
        queue.enqueueCopyBufferToImage(processedVolume, imageVolume, 0, offset, region);

        // Run blurVolumeWithGaussian on processedVolume (again)
        blurVolumeWithGaussianKernel.setArg(0, imageVolume);
        blurVolumeWithGaussianKernel.setArg(1, blurredVolume);
        blurVolumeWithGaussianKernel.setArg(2, 1.0f); // with higher sigma for GVF
        queue.enqueueNDRangeKernel(
                blurVolumeWithGaussianKernel,
                NullRange,
                NDRange(SIZE_X,SIZE_Y,SIZE_Z),
                NullRange
        );
    
        // Copy buffer to image
        queue.enqueueCopyBufferToImage(blurredVolume, imageVolume, 0, offset, region);
        
        // Run create vector field
        createVectorFieldKernel.setArg(0, imageVolume);
        createVectorFieldKernel.setArg(1, vectorField);
        createVectorFieldKernel.setArg(2, Fmax);

        queue.enqueueNDRangeKernel(
                createVectorFieldKernel,
                NullRange,
                NDRange(SIZE_X,SIZE_Y,SIZE_Z),
                NullRange
        );
        
        // Copy buffer to image
        queue.enqueueCopyBufferToImage(vectorField, iVectorField, 0, offset, region);
        

        // Run GVF on iVectorField as initial vector field
        Kernel GVFInitKernel = Kernel(program, "GVF3DInit");
        Kernel GVFIterationKernel = Kernel(program, "GVF3DIteration");
        Kernel GVFFinishKernel = Kernel(program, "GVF3DFinish");

        vectorField = Buffer(context, CL_MEM_READ_WRITE, 3*sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z);
        Buffer vectorField2 = Buffer(context, CL_MEM_READ_WRITE, 3*sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z);
        std::cout << "Running GVF... ( " << ITERATIONS << " )" << std::endl; 

        // init vectorField from image
        GVFInitKernel.setArg(0, iVectorField);
        GVFInitKernel.setArg(1, vectorField);
        queue.enqueueNDRangeKernel(
                GVFInitKernel,
                NullRange,
                NDRange(SIZE_X,SIZE_Y,SIZE_Z),
                NullRange
        );
        // Run iterations
        GVFIterationKernel.setArg(0, iVectorField);
        GVFIterationKernel.setArg(3, MU);

        for(int i = 0; i < ITERATIONS; i++) {
            if(i % 2 == 0) {
                GVFIterationKernel.setArg(1, vectorField);
                GVFIterationKernel.setArg(2, vectorField2);
            } else {
                GVFIterationKernel.setArg(1, vectorField2);
                GVFIterationKernel.setArg(2, vectorField);
            }
                queue.enqueueNDRangeKernel(
                        GVFIterationKernel,
                        NullRange,
                        NDRange(SIZE_X,SIZE_Y,SIZE_Z),
                        NullRange
                );
        }

        // Copy vector field to image
        vectorField2 = Buffer(context, CL_MEM_READ_WRITE, 4*sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z);
        GVFFinishKernel.setArg(0, vectorField);
        GVFFinishKernel.setArg(1, vectorField2);
        // TEST CODE
        Buffer sqrMag = Buffer(context, CL_MEM_WRITE_ONLY, sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z);
        GVFFinishKernel.setArg(2, sqrMag);
        // END TEST CODE
        queue.enqueueNDRangeKernel(
                GVFFinishKernel,
                NullRange,
                NDRange(SIZE_X,SIZE_Y,SIZE_Z),
                NullRange
        );
        Image3D imageVectorField2 = Image3D(context, CL_MEM_READ_ONLY, ImageFormat(CL_RGBA, CL_FLOAT), SIZE_X, SIZE_Y, SIZE_Z);
        queue.enqueueCopyBufferToImage(vectorField2, imageVectorField2, 0, offset, region);

        // TEST CODE
        float * F = new float[SIZE_X*SIZE_Y*SIZE_Z*3];
        //T.Fx = new float[SIZE_X*SIZE_Y*SIZE_Z];
        //T.Fy = new float[SIZE_X*SIZE_Y*SIZE_Z];
        //T.Fz = new float[SIZE_X*SIZE_Y*SIZE_Z];
        queue.enqueueReadBuffer(vectorField, CL_TRUE, 0, sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z*3, F);
        float * M = new float[SIZE_X*SIZE_Y*SIZE_Z];
        queue.enqueueReadBuffer(sqrMag, CL_TRUE, 0, sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z, M);
        /*
        for(int i = 0; i < SIZE_X*SIZE_Y*SIZE_Z; i ++) {
            //M[i] = sqrt(F[i*3]*F[i*3]+F[i*3+1]*F[i*3+1]+F[i*3+2]*F[i*3+2]);
            T.Fx[i] = F[i*3];
            T.Fy[i] = F[i*3+1];
            T.Fz[i] = F[i*3+2];
        }
        */
        writeToRaw<float>(F, "/home/smistad/gvf.raw", SIZE_X*3, SIZE_Y, SIZE_Z);
            
        writeToRaw<float>(M, "/home/smistad/sqrMag.raw", SIZE_X, SIZE_Y, SIZE_Z);
        free(M);
        free(F);
        // END TEST CODE

        std::cout << "Running TDF on GVF result..." << std::endl;

        // Run circle fitting TDF kernel on GVF result
        circleFittingTDFKernel.setArg(0, imageVectorField2);
        circleFittingTDFKernel.setArg(1, bufferT);
        circleFittingTDFKernel.setArg(2, bufferRadius);
        circleFittingTDFKernel.setArg(3, 1.0f);
        circleFittingTDFKernel.setArg(4, rMax);
        circleFittingTDFKernel.setArg(5, 1.0f);

        queue.enqueueNDRangeKernel(
                circleFittingTDFKernel,
                NullRange,
                NDRange(SIZE_X,SIZE_Y,SIZE_Z),
                NullRange
        );

        // Transfer buffer back to host
        float * Tlarge = new float[SIZE_X*SIZE_Y*SIZE_Z];
        float * Radiuslarge = new float[SIZE_X*SIZE_Y*SIZE_Z];
        queue.enqueueReadBuffer(bufferT, CL_TRUE, 0, sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z, Tlarge);
        queue.enqueueReadBuffer(bufferRadius, CL_TRUE, 0, sizeof(float)*SIZE_X*SIZE_Y*SIZE_Z, Radiuslarge);


        float * Tmerged = new float[SIZE_X*SIZE_Y*SIZE_Z];
        float * Radiusmerged = new float[SIZE_X*SIZE_Y*SIZE_Z];
        for(int i = 0; i < SIZE_X*SIZE_Y*SIZE_Z; i++) {
            if(Tsmall[i] < Tlarge[i]) {
                Tmerged[i] = Tlarge[i];
                Radiusmerged[i] = Radiuslarge[i];
            } else {
                Tmerged[i] = Tsmall[i];
                Radiusmerged[i] = Radiussmall[i];
            }
        }
        T.value = Tmerged;
        T.radius = Radiusmerged;
        free(Tlarge);
        free(Tsmall);
        free(Radiuslarge);
        free(Radiussmall);
        
        // Write buffer to RAW file
        std::cout << "Writing data to RAW file on disk..." << std::endl;
        writeToRaw<float>(Tmerged, "/home/smistad/output.raw", SIZE_X, SIZE_Y, SIZE_Z);
        writeToRaw<float>(Radiusmerged, "/home/smistad/radius.raw", SIZE_X, SIZE_Y, SIZE_Z);
    } catch(Error error) {
       std::cout << error.what() << "(" << error.err() << ")" << std::endl;
       std::cout << getCLErrorString(error.err()) << std::endl;
    }
    return T;
}
