#include "rawFiles.hpp"
#include <itkImage.h>
#include <itkRawImageIO.h>
#include <itkImageFileReader.h>
#include <itkImageRegionIterator.h>
#include <itkImageFileWriter.h>

short * parseRawFile(char * filename, int * SIZE_X, int * SIZE_Y, int * SIZE_Z) {
    itk::ImageIOBase::Pointer imageIO = itk::ImageIOFactory::CreateImageIO(filename, itk::ImageIOFactory::ReadMode);
imageIO->SetFileName(filename);
    imageIO->ReadImageInformation();



    /*
    itk::RawImageIO<short, 3>::Pointer io;
    io = itk::RawImageIO<short, 3>::New();
    io->SetFileName(filename);
    io->SetDimensions(0,SIZE_X);
    io->SetDimensions(1,SIZE_Y);
    io->SetDimensions(2,SIZE_Z);
    io->SetHeaderSize(0);
    io->SetByteOrderToLittleEndian();

    typedef itk::Image<short, 3> ImageType;
    itk::ImageFileReader<ImageType>::Pointer reader;
    reader = itk::ImageFileReader<ImageType>::New();
    reader->SetFileName(filename);
    reader->SetImageIO(io);
    reader->Update();
*/
    /*

    // Find min and max
    float min = 10;
    float max = -10;
    ImageType::Pointer im = reader->GetOutput();
    itk::ImageRegionIterator<ImageType> it(im, im->GetRequestedRegion() );
    for(it = it.Begin(); !it.IsAtEnd(); ++it) {
        if(it.Get() > max)
            max = it.Get();

        if(it.Get() < min)
            min = it.Get();

    }

    std::cout << "Min: " << min << " Max: " << max << std::endl;

    // Normalize result
    float * voxels = new float[SIZE_X*SIZE_Y*SIZE_Z];
    it = it.Begin();
    for(int i = 0; i < SIZE_X*SIZE_Y*SIZE_Z; i++) {
        voxels[i] = (float)(it.Get() - min) / (max - min);
        ++it;
    }*/
    typedef itk::Image<short, 3> ImageType;
    typedef itk::ImageFileReader<ImageType> ReaderType;
    ReaderType::Pointer reader = ReaderType::New();
    reader->SetFileName(filename);
    reader->Update();

    ImageType::Pointer im = reader->GetOutput();
    *SIZE_X = im->GetLargestPossibleRegion().GetSize()[0];
    *SIZE_Y = im->GetLargestPossibleRegion().GetSize()[1];
    *SIZE_Z = im->GetLargestPossibleRegion().GetSize()[2];
    short * voxels = new short[(*SIZE_X)*(*SIZE_Y)*(*SIZE_Z)];
    itk::ImageRegionIterator<ImageType> it(im, im->GetRequestedRegion() );
    it = it.Begin();
    int s = 0;
    if(imageIO->GetComponentType() == itk::ImageIOBase::USHORT) 
        s = 1024;
    for(int i = 0; i < (*SIZE_X)*(*SIZE_Y)*(*SIZE_Z); i++) {
        voxels[i] = it.Get()-s;
        ++it;
    }

    return voxels;
} 

short * lungCropping(short * dataset, int * SIZE_X, int * SIZE_Y, int * SIZE_Z) {
    short HUlimit = -150;
    int Wlimit = 30;
    int Blimit = 30;
    int minScanLines = 128;
    int x1,x2,y1,y2,z1,z2;
    z1 = *SIZE_Z / 2;
    z2 = *SIZE_Z / 2;

    for(int sliceDirection = 0; sliceDirection < 3; sliceDirection++) {
        int sliceSize, scanLineSize, scanLineElementSize;
        if(sliceDirection == 0) {
            sliceSize = *SIZE_X;
            scanLineSize = *SIZE_Y;
            scanLineElementSize = *SIZE_Z;
        } else if(sliceDirection == 1) {
            sliceSize = *SIZE_Y;
            scanLineSize = *SIZE_X;
            scanLineElementSize = *SIZE_Z;
        } else {
            sliceSize = *SIZE_Z;
            scanLineSize = *SIZE_Y;
            scanLineElementSize = *SIZE_X;
        }

        short * scanLinesInside = new short[sliceSize];

        for(int sliceNr = 0; sliceNr < sliceSize; sliceNr++) {
            scanLinesInside[sliceNr] = 0;
        for(int scanLine = 0; scanLine < scanLineSize; scanLine++) {
            int currentWcount = 0,
                currentBcount = 0,
                detectedBlackAreas = 0,
                detectedWhiteAreas = 0;
        for(int scanLineElement = 0; scanLineElement < scanLineElementSize; scanLineElement++) {
            int x,y,z;
            if(sliceDirection == 0) {
                x = sliceNr;
                y = scanLine;
                z = scanLineElement;
            }else if(sliceDirection == 1) {
                x = scanLine;
                y = sliceNr;
                z = scanLineElement;
            } else {
                x = scanLineElement;
                y = scanLine;
                z = sliceNr;
            }
            if(dataset[x+y*(*SIZE_X)+z*(*SIZE_X)*(*SIZE_Y)] > HUlimit) {
                if(currentWcount == Wlimit) {
                    detectedWhiteAreas++;
                    currentBcount = 0;
                }
                currentWcount++;
            } else {
                if(currentBcount == Blimit) {
                    detectedBlackAreas++;
                    currentWcount = 0;
                }
                currentBcount++;
            }
        } // End scan line element

        if((detectedWhiteAreas == 2 && detectedBlackAreas == 1) || 
                (detectedBlackAreas > 1 && detectedWhiteAreas > 1))
            scanLinesInside[sliceNr]++;
        } // End scan line
        } // End slice

        if(sliceDirection == 2) {
            for(int sliceNr = sliceSize/2; sliceNr < sliceSize; sliceNr++) {
                if(scanLinesInside[sliceNr] < minScanLines) {
                    z2 = sliceNr;
                    break;
                }
            }
            for(int sliceNr = sliceSize/2; sliceNr > 0; sliceNr--) {
                if(scanLinesInside[sliceNr] < minScanLines) {
                    z1 = sliceNr;
                    break;
                }
            }
        } else {
             for(int sliceNr = 0; sliceNr < sliceSize; sliceNr++) {
                if(scanLinesInside[sliceNr] > minScanLines) {
                    if(sliceDirection == 0) {
                        x1 = sliceNr;
                    } else {
                        y1 = sliceNr;
                    }
                    break;
                }
            }
            for(int sliceNr = sliceSize-1; sliceNr > 0; sliceNr--) {
                if(scanLinesInside[sliceNr] > minScanLines) {
                    if(sliceDirection == 0) {
                        x2 = sliceNr;
                    } else {
                        y2 = sliceNr;
                    }
                    break;
                }
            }       
        }
    } // End for each direction

    int newSIZE_X, newSIZE_Y, newSIZE_Z;
    newSIZE_X = x2-x1;
    newSIZE_Y = y2-y1;
    newSIZE_Z = z2-z1;

    short * croppedDataset = new short[newSIZE_X*newSIZE_Y*newSIZE_Z];
    for(int x = x1; x < x2; x++) {
        for(int y = y1; y < y2; y++) {
            for(int z = z1; z < z2; z++) {
                croppedDataset[x-x1+(y-y1)*newSIZE_X+(z-z1)*newSIZE_X*newSIZE_Y] 
                    = dataset[x+y*(*SIZE_X)+z*(*SIZE_X)*(*SIZE_Y)]; 
            }
        }
    }
    *SIZE_X = newSIZE_X;
    *SIZE_Y = newSIZE_Y;
    *SIZE_Z = newSIZE_Z;
    return croppedDataset;
}
